/* GNU POP3 - a small, fast, and efficient POP3 daemon
   Copyright (C) 1999 Jakob 'sparky' Kaivo <jkaivo@nodomainname.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "gnu-pop3d.h"

#ifdef MDN_LANGUAGE_EN
#define MDN_MSG "The message\r\n\r\n  To: %s\r\n  From: %s\r\n  Subject: %s\r\n  Sent: %s\r\n\r\nwas retrieved (not necessarily read) on %s."
#define MDN_WARN "Warning: The message has been retrieved, but not necessarily read\r\n\r\n"
#endif /*  */

#ifdef MDN_LANGUAGE_DA
#define MDN_MSG "Beskeden\r\n\r\n  Til: %s\r\n  Fra: %s\r\n  Emne: %s\r\n  Sendt: %s\r\n\r\nblev modtaget (ikke n=F8dvendigvis l=E6st) %s."
#define MDN_WARN "OBS: Beskeden blev modtaget, men muligvis ikke laest"
#endif /*  */

#include <unistd.h>

/* Sends a Message Disposition Notification to the sender */
int
pop3_sendmdn (char *returnpath, char *envto, char *delivto, char *subject,
			  char *date, char *msgid, char *sender)
{
  FILE *fil;
  int fd, rc = 0;
  char hostname[64];
  char *tempfile1 = "/GNUpop3dXXXXXX";
  char *tempfile = malloc (strlen (tmppath) + strlen (tempfile1) + 1);
  char cmd[256];
  char currdate[64];
  time_t now = time (NULL);
  char *p, *p1, *tmp, *tmp1, *receiv;

  if (debug & 0x20)
	syslog (LOG_INFO,
			"pop3_sendmdn: mdn=%d sender=%s returnpath=%s envto=%s delivto=%s",
			mdn, sender, returnpath, envto, delivto);

  if (mdn == 0)
	return OK;
  if (envto == NULL)
	receiv = delivto;
  else
	receiv = envto;
  if ((sender == NULL) || (returnpath == NULL) || (receiv == NULL))
	return ERR_MDN;

/* don't send if more than 1 receiver or no return path */
  if ((index (sender, ',') != NULL) || (strstr (returnpath, "<>") != NULL))
	return ERR_MDN;

/* RFC2298: check that return path address <...> and sender <...> are equal, case-sensitive before @ */
  p = strchr (sender, '<');
  if (p == NULL)
	tmp = strdup (sender);

  else
	tmp = strdup (++p);
  if (tmp != NULL)
	{
	  if ((p1 = strchr (tmp, '>')) != NULL)
		*p1 = '\0';
	}
  else
	return ERR_MDN;
  p = strchr (returnpath, '<');
  if (p == NULL)
	tmp1 = strdup (returnpath);

  else
	tmp1 = strdup (++p);
  if (tmp1 != NULL)
	{
	  if ((p1 = strchr (tmp1, '>')) != NULL)
		*p1 = '\0';
	}
  else
	return ERR_MDN;
  if (strcasecmp (tmp, tmp1) != 0)
	return ERR_MDN;

  else
	{
	  if ((p = strchr (tmp, '@')) != NULL)
		*p = '\0';
	  if ((p = strchr (tmp1, '@')) != NULL)
		*p = '\0';
	  if (strcmp (tmp, tmp1) != 0)
		return ERR_MDN;
	}
  strftime (currdate, 64, "%a, %d %b %Y %H:%M:%S (%Z)", localtime (&now));
  tempfile = strcpy (tempfile, tmppath);
  strcat (tempfile, tempfile1);

/* create message according to RFC 2298 */
  fd = mkstemp (tempfile);
  snprintf (cmd, 255, MDN_MAILCMD, tempfile);

  if (debug & 0x20)
	syslog (LOG_INFO, "pop3_sendmdn: mdn=%d opt=%s %s tmp=%s cmd=%s", mdn,
			useroptions, currdate, tempfile, cmd);

  if (fd != -1)
	{
	  fil = fdopen (fd, "w");
	  fprintf (fil, "To: %s\r\n", sender);
	  fprintf (fil, "Subject: Re: %s\r\n", (subject == NULL) ? "" : subject);
	  fprintf (fil, "From: <>\r\n");
	  fprintf (fil, "Cc: %s\r\n", receiv);
	  fprintf (fil, "MIME-Version: 1.0\r\n");
	  fprintf (fil,
			   "Content-Type: multipart/report;\r\n\treport-type=disposition-notification;\r\n\tboundary=\"----=_NextPart_000_0001_01C15828.00615740\"\r\n\r\n");
	  fprintf (fil, "This is a multi-part message in MIME format.\r\n\r\n");
	  fprintf (fil, "------=_NextPart_000_0001_01C15828.00615740\r\n");
	  fprintf (fil, "MIME-Version: 1.0\r\n");
	  fprintf (fil, "Content-Type: text/plain; charset=iso-8859-1\r\n");
	  fprintf (fil, "Content-Transfer-Encoding: quoted-printable\r\n\r\n");
	  fprintf (fil, MDN_MSG, receiv, sender, subject, date, currdate);
	  fprintf (fil, "\r\n\r\n");
	  fprintf (fil, "------=_NextPart_000_0001_01C15828.00615740\r\n");
	  fprintf (fil, "Content-Type: message/disposition-notification\r\n");
	  fprintf (fil, "Content-Transfer-Encoding: 7bit\r\n\r\n");
	  gethostname (hostname, 63);
	  fprintf (fil, "Reporting-UA: %s;%s-%s\r\n", hostname, IMPL, VERSION);
	  fprintf (fil, "Final-Recipient: rfc822;%s\r\n", receiv);
	  fprintf (fil, "Original-Message-ID: %s\r\n", msgid);
	  fprintf (fil, "Disposition: automatic-action/MDN-sent-automatically;dispatched\r\n");
	  fprintf (fil, MDN_WARN);
	  fprintf (fil, "\r\n\r\n");
	  fprintf (fil, "------=_NextPart_000_0001_01C15828.00615740--\n");
	  fclose (fil);

/* send it by calling external mailprogram */
	  if ((rc = system (cmd)) != 0)
		syslog (LOG_INFO, "pop3_sendmdn: send MDN failed with rc=%d cmd=%s", rc, cmd);

	  else
		syslog (LOG_INFO, "Sent MDN to %s from %s", sender, receiv);
	  remove (tempfile);
	}
  else
	syslog (LOG_INFO, "pop3_sendmdn: could not create %s rc=%d", tempfile, rc);
  free (tempfile);
  return OK;
}


/* search for specific string in header line prefix and return data from the header
   eg. pop3_pickupheader(headerline, "X-Envelope-to:")
   on X-Envelope-To: abc@def.com will return abc@def.com */
char *
pop3_pickupheader (const char *haystack, const char *needle)
{
  char *res = NULL;
  char *h, *n, *p, *c;
  h = strdup (haystack);
  n = strdup (needle);
  if ((h == NULL) || (n == NULL))
	return res;
  p = h;						/* lower case haystack */
  while (*p != '\0')
	{
	  *p = tolower (*p);
	  p++;
	}
  p = n;						/* lower case needle */
  while (*p != '\0')
	{
	  *p = tolower (*p);
	  p++;
	}
  c = strchr (h, ':');
  if (c != NULL)
	{
	  *c = '\0';
	  p = h;					/* strip leading space */
	  while (p != c && isgraph (*p) == 0)
		{
		  p++;
		}

	  if (strcmp (p, n) == 0)	/* is headerline id found */
		{
		  c = strchr (haystack, ':');
		  c++;
		  while (isgraph (*c) == 0)
			c++;
		  res = strdup (c);
		  p = index (res, '\n');
		  if (p != NULL)
			*p = '\0';
		  p = index (res, '\r');
		  if (p != NULL)
			*p = '\0';
		}
	}

  if ((debug & 0x88) == 0x88)
	syslog (LOG_INFO, "pop3_pickupheader: find %s in %s gave %s", needle, haystack, res);

  free (h);
  free (n);
  return res;
}
