/* GNU POP3 - a small, fast, and efficient POP3 daemon
   Copyright (C) 1999 Jakob 'sparky' Kaivo <jkaivo@nodomainname.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "gnu-pop3d.h"
#include <unistd.h>

#define LINESLICE 160

int pop3_quit_newfile (char *mailbox);
int pop3_quit_oldfile (fpos_t * readpos, fpos_t * writepos);
int pop3_copy_msg (const int mesg, FILE * mbox, FILE * newmbox);

/* Enters the UPDATE phase and deletes marked messages */

int
pop3_quit (const char *arg)
{
  int i, count;
  fpos_t *readpos = NULL, *writepos = NULL;
  int rc = 0;
  if (strlen (arg) != 0)
	return ERR_BAD_ARGS;

  if (state == TRANSACTION)
	{
	  count = 0;
	  for (i = 0; i < num_messages; i++)
		if ((messages[i].chng) && !messages[i].deleted)
		  count++;				/* msg disp notif sent or new msg retrieved */

	  if (count > 0)			/* Status: R or O must be written to header */
		rc = pop3_quit_newfile (mailbox);
	  else
		{
		  readpos = malloc (sizeof (fpos_t));
		  writepos = malloc (sizeof (fpos_t));
		  if (readpos == NULL || writepos == NULL)
			pop3_abquit (ERR_NO_MEM);
		  rc = pop3_quit_oldfile (readpos, writepos);
		  free (readpos);
		  free (writepos);
		}
	  free (messages);
	  messages = NULL;
	  syslog (LOG_INFO, "Session ended for user: %s", username);
	}
  else
	syslog (LOG_INFO, "Session ended for no user");

  state = UPDATE;
/*  if (username)
	{
	  free (username);
	  username = NULL;
	} */
  if (mailbox)
	{
	  free (mailbox);
	  mailbox = NULL;
	}
  free (md5shared);
  md5shared = NULL;
  if (rc == 0)
	{
	  fprintf (ofile, "+OK\r\n");
	  rc = OK;
	}
  else
	rc = ERR_FILEPOS;
  if (rc != 0 || debug & 0x80)
	syslog (LOG_INFO, "QUIT rc = %i", rc);
  return rc;
}


int
pop3_copy_msg (const int mesg, FILE * mbox, FILE * ofile)
{
  int empty_line = 0;
  int lbuf, done = 0, chwrt = 0, charcount = 0, linecount = 0;
  char buf[LINESLICE * sizeof (char)], lineend[3 * sizeof (char)];
  int header = 1;
  char *res;

  if (pop3_mesg_exist (mesg) != OK)
	return ERR_NO_MESG;

  if (fsetpos (mbox, &(messages[mesg].msgstart)) < 0)
	return ERR_FILEPOS;

  while ((res = fgets (buf, LINESLICE, mbox)) && !done)
	{
	  if (empty_line && !strncmp (buf, "From ", 5))
		done = 1;
	  else
		{
		  lineend[0] = '\0';
		  if ((buf[0] == '\n') || (buf[0] == '\r') || (buf[0] == '\0'))
			empty_line = 1;
		  else
			empty_line = 0;
		  lbuf = strlen (buf);
		  if (buf[lbuf - 1] == '\n')
			{
			  linecount++;
			  lbuf--;
			  strcpy (lineend, "\n");
			}
		  if (header && messages[mesg].chng && (empty_line || strncmp ("Status:", buf, 7) == 0))
			{
			  if (empty_line)
				{
				  header = 0;
				  errno = 0;
				  if ((chwrt = fprintf (ofile, "Status: %s\n\n", messages[mesg].status)) < 10)
					return ERR_FILE;
				  if (errno)
					{
					  syslog (LOG_INFO, "Error %d on new mailbox (a): %s", errno, strerror (errno));
					  return errno;
					}
				  linecount += 2;
				  charcount += chwrt;
				}
			  else				/* skip old Status: line */
				while ((res != NULL) && (strchr (buf, '\n') == NULL))
				  res = fgets (buf, LINESLICE, mbox);
			}
		  else
			{
			  errno = 0;
			  if ((chwrt = fprintf (ofile, "%.*s%s", lbuf, buf, lineend)) < lbuf + strlen (lineend))
				return ERR_FILE;
			  if (errno)
				{
				  syslog (LOG_INFO, "Error %d on new mailbox (b): %s", errno, strerror (errno));
				  return errno;
				}
			  charcount += chwrt;

			  while ((res != NULL) && (lineend[0] == '\0' && strchr (buf, '\n') == NULL))
				{
				  res = fgets (buf, LINESLICE, mbox);
				  if (res == NULL)
					{
					  strcpy (buf, "\n\n");
					}			/* last line in mailbox without \n */
				  lbuf = strlen (buf);
				  if (buf[lbuf - 1] == '\n')
					{
					  linecount++;
					  lbuf--;
					  strcpy (lineend, "\n");
					}
				  errno = 0;
				  if ((chwrt =
					   fprintf (ofile, "%.*s%s", lbuf, buf, lineend)) < lbuf + strlen (lineend))
					return ERR_FILE;
				  if (errno)
					{
					  syslog (LOG_INFO, "Error %d on new mailbox (c): %s", errno, strerror (errno));
					  return errno;
					}
				  charcount += chwrt;
				}
			}
		}
	}
/* printf("num: %d pos: %ld, %ld ch: %d %s sizes=%d %d lc=%d\n", mesg, messages[mesg].header.__pos, messages[mesg].msgstart.__pos, messages[mesg].chng, messages[mesg].status, messages[mesg].size, charcount, linecount); */
  return OK;
}

int
pop3_quit_newfile (char *mailbox)
{
  int mesg, rc = 0, count = 0;
  FILE *newmbox = NULL;
  char *newmailbox = NULL;
  char *oldmailbox = NULL;

  if (num_messages > 0)
	{							/* mailbox non-empty or not wrong format */
	  oldmailbox = malloc (sizeof (char) * (strlen (mailbox) + 14));
	  newmailbox = malloc (sizeof (char) * (strlen (mailbox) + 14));	/* 4 is not enough to avoid signal 11 at free !! */
	  if (newmailbox == NULL)
		return ERR_NO_MEM;
	  strcpy (newmailbox, mailbox);
	  strcat (newmailbox, ".new");
	  strcpy (oldmailbox, mailbox);
	  strcat (oldmailbox, ".old");
	  syslog (LOG_INFO, "Copying mailbox to %s. %d MDNs sent", newmailbox, mdncount);

	  if (fseek (mbox, 0L, SEEK_SET))
		return ERR_FILE;
	  newmbox = fopen (newmailbox, "w");
	  if (newmbox == NULL)
		rc = ERR_FILE;

	  for (mesg = 0; (mesg < num_messages) && (rc == 0); mesg++)
		{
		  rc = pop3_copy_msg (mesg, mbox, newmbox);
		  if (!rc)
			count++;
		  if (rc == ERR_NO_MESG)
			rc = 0;
		}
	  if (newmbox)
		{
		  if (fsync (fileno (newmbox)) == -1)
			rc = ERR_FILE;
		  fclose (newmbox);
		}
	  if (!rc)
		{

		  rc = rename (mailbox, oldmailbox);
		  if (rc)
			{
			  syslog (LOG_INFO, "ERROR: %s not renamed to %s RC=%d", mailbox, oldmailbox, rc);
			  rc = ERR_FILE;
			}
		  else
			{
			  rc = ftruncate (fileno (mbox), 0);
			  if (rc)
				{
				  syslog (LOG_INFO, "ERROR: %s not truncated RC=%d", oldmailbox, rc);
				}
			  rc = rename (newmailbox, mailbox);
			  syslog (LOG_INFO,
					  "%d of %d copied. %d deleted. Renaming %s to %s RC=%d",
					  count, num_messages, cursor, newmailbox, mailbox, rc);
			  if (rc)
				{
				  rc = ERR_FILE;
				}
			}
		}
	  else
		unlink (newmailbox);
	}
  pop3_unlock ();
  free (newmailbox);
  newmailbox = NULL;
  mbox = NULL;
  return rc;
}

int
pop3_quit_oldfile (fpos_t * readpos, fpos_t * writepos)
{
  int i = 0, size = 0, write = 0, rc = 0, chwrt;
  int new_line = 1, empty_line = 1;
  char buf[LINESLICE * sizeof (char)];
  FILE *ombox;
  char *res;

  ombox = fopen (mailbox, "r+");
  if (ombox == NULL)
	return ERR_FILE;
  if (fseek (mbox, 0L, SEEK_SET))
	return ERR_FILE;
  if (fgetpos (mbox, writepos) < 0)
	return ERR_FILEPOS;

  if (debug & 0x20)
	syslog (LOG_INFO, "Total messages = %d Messages to delete = %d", num_messages, cursor);

  if ((cursor > 0) && (num_messages > 0))
	{
	  if (num_messages == cursor)
		{
		  ftruncate (fileno (mbox), 0);	/* quickly delete everything */
		  if (debug & 0x20)
			syslog (LOG_INFO, "Deleted all messages");
		}
	  else
		{
		  while ((res = fgets (buf, LINESLICE, mbox)) && (rc == 0))
			{
			  if (empty_line && !strncmp (buf, "From ", 5))
				{
				  if (pop3_mesg_exist (i) == OK)
					write = 1;
				  else
					write = 0;
				  i++;
				}
			  if (write)
				{
#if 0
				  rc = fgetpos (mbox, readpos);
				  if (rc == 0)
					{
					  rc = fsetpos (mbox, writepos);
					  if (rc == 0)
						{
						  chwrt = fprintf (mbox, "%s", buf);
						  if (chwrt != strlen (buf))
							rc = ERR_FILE;
						  size += chwrt;
						  rc = fgetpos (mbox, writepos);
						  if (rc == 0)
							rc = fsetpos (mbox, readpos);
						}
					}
#else
				  /* much faster */
				  errno = 0;
				  chwrt = fprintf (ombox, "%s", buf);
				  if (chwrt != strlen (buf))
					rc = ERR_FILE;
				  if (errno)
					{
					  syslog (LOG_INFO, "Error %d on mailbox: %s", errno, strerror (errno));
					  return errno;
					}
				  size += chwrt;
#endif
				}
			  /* "From " only separates messages if followed by blank line */
			  if (new_line && (buf[0] == '\n' || buf[0] == '\r'))
				empty_line = 1;
			  else
				empty_line = 0;
			  if (strchr (buf, '\n'))
				new_line = 1;
			  else
				new_line = 0;
			}

		  if ((rc == 0) && (new_line == 0))
			{
			  fprintf (ombox, "%s", "\n\n");
			  size += 2;
			}

		  fclose (ombox);
		  if (rc == 0)
			ftruncate (fileno (mbox), sizeof (char) * size);
		}
	}
  pop3_unlock ();
  fclose (mbox);
  mbox = NULL;
  return rc;
}
